

# This file contains the default settings for the application. First we look in the 
# global application settings.py file for any settings that are defined there. Then
# we set any default values.

# To use these settings, you must:
# from .settings import AUTH

# If you instead defer to the settings.py file, you will not have the default values
# which are defined in this file.

from django.conf import settings


defaults = {
    # TODO: Change require activation to False
    'REQUIRE_ACTIVATION': True,
    'RESET_PASSWORD_EXPIRES': 20,
    'REQUIRE_PASSWORD_ON_ACCOUNT_MODIFICATIONS': True,
    'USERNAME_FIELD': 'email',
    'LOGIN_WITH_FIELDS': ['email'],
    'LOGIN_WITH_PARAM': 'username',
    'EMAIL_REQUIRED': True,
    'REQUIRED_FIELDS': [],

    'RESET_PASSWORD_REQUEST_ALWAYS_OK': True,

    # TODO: Implement open registration
    'OPEN_REGISTRATION': True,

    'FROM_EMAIL_ADDRESS': 'admin'
}


# get the section from the app settings.py file
AUTH = getattr(settings, 'AUTH', {})

# set the defaults defined above
for key, value in defaults.items():
    AUTH[key] = AUTH.get(key, defaults[key] )

# modify the "LOGIN_WITH_FIELDS" array to always include the USERNAME_FIELD
if not AUTH['USERNAME_FIELD'] in AUTH['LOGIN_WITH_FIELDS']:
    AUTH['LOGIN_WITH_FIELDS'] = AUTH['LOGIN_WITH_FIELDS'].insert(0, AUTH['USERNAME_FIELD'] )




from django.test import TestCase
from rest_framework.test import APIRequestFactory
from rest_framework.test import APIClient
import json
from django.utils import timezone
from datetime import timedelta

# Create your tests here.
from .models import User, UserManager, UserActivation, ResetPasswordRequest
from .serializers import UserSerializer

from .settings import AUTH, defaults as AUTH_DEFAULTS

from agape.auth import username_opt, username_value, username_field
AUTH['USERNAME_FIELD']="username"

import copy
from agape.test import AgapeTests, ApiMixin, ApiTests, ModelTests, SerializerTests, TestCaseMixin
from django.conf import settings



from .test_default import USER_MODEL, USER_SERIALIZER, USER_ENTITY_NAME, USER_API_END_POINT

USER_ENTITY_NAME = 'user'
USER_API_END_POINT = '/api/users/'
AUTH_API_END_POINT = '/api/auth/'
REGISTER_API_END_POINT = "/api/register"
RECOVERY_API_END_POINT = "/api/auth/recovery/"

USER_CREATE_DATA = [

    {
        'username': 'ajones',
        'password': 'testpass',
        'email': 'testing@example.com'
    },
    {
        'username': 'jsmith',
        'password': 'foobar12',

    }
]


USER_UPDATE_DATA = [
    
    {
        'username': 'achandler'
    },
    {
        'password': 'barfood12'
    }

]

class UserTestMixin( TestCaseMixin ):

    def sharedSetUp(self):
        self.resetSettings()

        self.model              = USER_MODEL
        self.serializer_class   = USER_SERIALIZER
        self.entity             = USER_ENTITY_NAME
        self.api_end_point      = USER_API_END_POINT
        self.auth_end_point     = AUTH_API_END_POINT
        self.register_end_point = REGISTER_API_END_POINT
        self.recovery_end_point = RECOVERY_API_END_POINT

        self.create_data        = copy.deepcopy(USER_CREATE_DATA)
        self.expect_data        = copy.deepcopy(USER_CREATE_DATA)
        for record in self.expect_data:
            record.pop('password')
            
        self.update_data        = copy.deepcopy(USER_UPDATE_DATA)
        self.update_expect_data = copy.deepcopy(USER_UPDATE_DATA)
        for record in self.update_expect_data:
            record.pop('password', None)

    def resetSettings(self):
        # reset configuration to default settings
        for key, value in AUTH_DEFAULTS.items():
            AUTH[key] = AUTH.get(key, AUTH_DEFAULTS[key] )

    def register_user( self, data ):
        response = self.client.post( self.register_end_point, data )
        self.assertEqual(response.status_code, 201, "Registered user")
        return response


class ModelTestCase( UserTestMixin, ModelTests, TestCase ):
    pass

class ModelSerializerCase( UserTestMixin, SerializerTests, TestCase ):
    pass



class APITestCase(UserTestMixin, ApiMixin, AgapeTests, TestCase):

    def setUp(self):
        self.sharedSetUp()
        self.createRequisiteObjects();
        self.serializedSetUp()
        self.maxDiff = None
        self.client =  APIClient()

    def test_user_registration(self):

        i=0
        while i < len(self.serialized_create_data):

            cdata = self.create_data[i]
            response = self.register_user( cdata )
            self.assertEqual( response.status_code, 201, "Registered new user")

            # verify the response data
            self.assertSerializedInstanceEqual( response.data, self.serialized_expect_data[i] )

            self.assertEqual( response.data['username'], cdata['username'] )

            # verify actual database record was created
            instance = self.model.objects.get(id=response.data.get('id'))
            self.assertTrue(instance)

            i+=1

    def test_login(self):

        cdata = self.create_data[0]
        instance = self.model.objects.create( **cdata, status=1 )

        # login
        response = self.authenticate( username=cdata['username'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")
        self.assertTrue(response.data.get('token'), "Received auth token.")


    def test_login_with_email(self):

        cdata = self.create_data[0]

        instance = self.model.objects.create( **cdata, status=1 )

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")
        self.assertTrue(response.data.get('token'), "Received auth token.")


    def test_login_with_email_as_username(self):


        cdata = self.create_data[0]

        instance = self.model.objects.create( **cdata, status=1 )

        # login
        response = self.authenticate( username=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")
        self.assertTrue(response.data.get('token'), "Received auth token.")



from django.conf.urls import url, include
from .views import UserViewSet, AuthView, ResetPasswordRequestView, ValidateEmailView


from .views import RegisterView
from .views import SendUserActivationView

# build router
from rest_framework.routers import DefaultRouter

from rest_framework_jwt.views import obtain_jwt_token, refresh_jwt_token, verify_jwt_token

router = DefaultRouter()
router.register(r'users', UserViewSet, basename='users')

urlpatterns = [
    url(r'^',include(router.urls)),

    url(r'^auth/$', AuthView.as_view(), name='auth'),
    url(r'^auth/recovery/((?P<key>[A-Z0-9_]+)/)?$', ResetPasswordRequestView.as_view(), name='reset-password'),

    url(r'^register/?$'  , RegisterView.as_view() ),
    
    url(r'^auth/email/?$', ValidateEmailView.as_view(), name='validate-email'),

    url(r'^auth/token-auth/', obtain_jwt_token),
    url(r'^auth/token-verify/', verify_jwt_token),
    url(r'^auth/token-refresh/', refresh_jwt_token),


    url(r'^users/(?P<pk>[0-9]+)/send-activation-email/', SendUserActivationView.as_view(), name="send-activation-email"),
]

    # users/
    #   - GET: list
    #   - POST: register
    #   - PATCH: update
    
    # auth/
    #   - POST: login
    #   - DELETE: logout
    #   - PATCH: activate account

    # auth/password
    #   - POST: create password reset request
    #   - PATCH: modify password
        
    # auth/email
    #   - GET: verify email
    #       - parameters:
    #           email
    #           userid


from rest_framework import permissions,status,views,viewsets
from rest_framework.response import Response
from rest_framework import generics

from django.core.mail import send_mail
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth import get_user_model
User = get_user_model()

from django.contrib.sites.shortcuts import get_current_site
from django.db import transaction
from rest_framework_jwt.settings import api_settings



from django.utils import timezone
from datetime import timedelta

from .permissions import IsAccountOwner
from .serializers import UserSerializer
from .models import UserActivation, ResetPasswordRequest
from .settings import AUTH

from agape import auth
from agape.auth import username_field, username_opt, username_value
from agape.proxies import ModelProxy
from agape.signals import trigger
from agape.viewsets import CrudViewSet

# update the session auth hash when modifying a logged in users credentials
from django.contrib.auth import update_session_auth_hash


class RegisterView( views.APIView ): # , ScopedViewMixin

	# context = "user"
	# serializer_class = UserSerializer

	def post( self, request, *args, **kwargs ):

		proxy = ModelProxy( ['user', 'user.register'], User, UserSerializer )

		proxy.create( **kwargs, request=request )

		return proxy.response


		
class UserViewSet( CrudViewSet ):

	context = "user"
	model = User
	serializer_class = UserSerializer

	
	def get_permissions(self):
		""" Set permission restrictions """

		# allow super admin to access all methods
		if self.request.user.is_superuser:
			return(permissions.AllowAny(),)

		if self.action == "list":
			return (permissions.IsAdminUser(),)

		else:
			return(permissions.IsAuthenticated(), IsAccountOwner() )

			
class SendUserActivationView( views.APIView ):

	def get_permissions(self):

		if self.request.user.is_superuser:
			return(permissions.AllowAny(),)
		else:
			return (permissions.IsAdminUser(),)

	def get( self, request, *args, **kwargs ):

		userid = kwargs.get('pk')

		user = User.objects.get( pk=userid )

		instance = UserActivation.objects.create( user=user )

		instance.send_activation_email( request )

		return Response('ok')

	



class AuthView(views.APIView):
	""" Handles auth (logging in and out)
	"""

	serializer_class = UserSerializer

	def post(self, request, format=None):
		""" Sign in to a user account .

		Returns a response with an auth token
		"""


		data = request.data
		user = None

		# TODO: Maybe LOGIN_WITH_PARAM should only make one call to the database
		# by using an OR query

		# Find the user, using all the possible login fields
		for field in AUTH['LOGIN_WITH_FIELDS']:
			if field in data:
				login_with_value = data.get(field)

			# the login_with_field can be used to sign in to any of the 'LOGIN_WITH_FIELDS' fields
			elif AUTH['LOGIN_WITH_PARAM']:
				login_with_value = data.get( AUTH['LOGIN_WITH_PARAM'] )

			try:
				lookup_opt = { field: login_with_value }
				user = User.objects.get( **lookup_opt )
			except:
				pass



		# return unauthorized response if the user was not found
		if user == None:
			return Response({
				'status': 'unauthorized',
				'message': 'Invalid credentials.'
			}, status=status.HTTP_401_UNAUTHORIZED)            
		
		# return unauthorized if the user is disabled
		if not user.is_active:
			return Response({
				'status': 'unauthorized',
				'message': 'This user has been disabled.'
			}, status=status.HTTP_401_UNAUTHORIZED)


		# we found a matching user, authenticate
		password = data.get('password', None)
		user = authenticate( **username_opt( username_value(user) ), password=password)


		# invalid credentials
		if user == None:
			return Response({
				'status': 'unauthorized',
				'message': 'Invalid credentials.'
			}, status=status.HTTP_401_UNAUTHORIZED)   


		# user authenticated successfully
		else:

			# login
			login(request, user)
			user.last_login = timezone.now()
			user.save()

			# create an auth token
			jwt_payload_handler = api_settings.JWT_PAYLOAD_HANDLER
			jwt_encode_handler = api_settings.JWT_ENCODE_HANDLER
			
			payload = jwt_payload_handler(user)
			token = jwt_encode_handler(payload)

			# create the response
			response = {
				'user': self.serializer_class(user).data,
				'token': token 
				}
			return Response( response, status=status.HTTP_200_OK );



	def delete(self, request, *args, **kwargs):
		""" Destroy a users auth token, effectively signing out.
		"""

		# If a user is signed in
		if request.user.is_authenticated:
			# sign out
			logout(request)
			return Response({}, status=status.HTTP_204_NO_CONTENT) 
		
		# otherwise respond with bad request
		else:
			# Do something for anonymous users.
			return Response({
				'status': 'Bad request',
				'message': 'There is no authenticated user to sign out.'
			}, status=status.HTTP_400_BAD_REQUEST)  


	def patch(self,request,*args,**kwargs):
		""" Activate a user account
		"""

		key = request.data.get('key')
		if key == None:
			return Response({
					'status': 'fail',
					'message': 'No activation key given.',
				},  status=status.HTTP_400_BAD_REQUEST);

		# perform key lookup
		results = UserActivation.objects.filter(key=key).all()

		# no activation key found in database
		if len(results) == 0:
			return Response({
					'status': 'fail',
					'message': 'Could not activate account',
				},  status=status.HTTP_400_BAD_REQUEST);

		# user has been banned
		# TODO: Replace this user status check with enumerated package variable
		elif results[0] and results[0].user.status < 0:
			return Response({
					'status': 'fail',
					'message': 'This account has been disabled.',
				},  status=status.HTTP_400_BAD_REQUEST);              

		# found activation key, but user is already activated
		elif results[0] and results[0].activated:
			return Response({
					'status': 'fail',
					'message': 'This account has already been activated.',
				},  status=status.HTTP_409_CONFLICT);   


		# found activation key, activate account
		else:
			activation = results[0]
			activation.activated = timezone.now()
			activation.save()

			activation.user.status = 1;
			activation.user.save();
			return Response({
					'status': 'succes',
					'message': 'This account is now active.',
				},  status=status.HTTP_200_OK); 



	

class ResetPasswordRequestView(views.APIView):
	""" Send a password reset link to a user via email
	
	Will send a password reset link if no other request has been issued in the 
	past X minutes. Password reset links are only valid for the number of minutes
	specified in the configuration file. Defaults to 30.

	Extends:
		views.APIView    
	"""
	@transaction.atomic
	def post(self,request,*args, **kwargs):
		""" Create a new reset password request. 

		Generates a new unique key which can be used to reset a users password. Sends
		an email containing a password reset link to the specified user.
		"""

		# TODO: Get the user via username or password
		# get key from uri
		email = request.data.get('email')

		# find user with given email
		user = User.objects.filter(email=email).all()

		# user does not exist- return response
		if len(user) == 0:

			if AUTH['RESET_PASSWORD_REQUEST_ALWAYS_OK']:

				return Response({
						'status': 'succes',
						'message': 'A link to reset your password has been sent to this email address.',
					},  status=status.HTTP_200_OK); 

			else:
				
				return Response({
						'status': 'fail',
						'message': 'No user with this email address.',
					},  status=status.HTTP_404_NOT_FOUND);


		# user does exist
		else:
			user = user[0]

		# if user has been disabled, throw an error
		if user.status == User.DISABLED:
			return Response({
					'status': 'fail',
					'message': 'This user has been disabled.',
				},  status=status.HTTP_400_BAD_REQUEST);
		
		# check if there is another password reset request sent in the past 30 minutes
		time_threshold = timezone.now() - timedelta(minutes=AUTH.get('RESET_PASSWORD_EXPIRES'))
		results = ResetPasswordRequest.objects.filter(user=user,status=ResetPasswordRequest.ENABLED, created__gt=time_threshold).order_by('-id')
				
		# if active request, use the same request link
		if len(results) > 0:
			instance = results[0]

		# if no active request, create a new request link
		else:
			instance = ResetPasswordRequest.objects.create(user=user)

		# create password reset instance
		self.dispatch_email(instance, request)

		return Response({
				'status': 'succes',
				'message': 'A link to reset your password has been sent to this email address.',
			},  status=status.HTTP_200_OK); 




	def get(self,request,*args, **kwargs):
		""" Check a reset password request for valididy and expiration
		"""

		# get key from uri
		key = self.kwargs.get('key')

		# check key for validity
		response = self.validate_key(key)

		# if received Response, invalid key
		if isinstance(response, Response):
			# return the response
			return response
		# otherwise, confirm password reset link is valid
		else:
			return Response({
					'status': 'success',
					'message': 'Password reset key is valid.'
				}, status=status.HTTP_200_OK);


	@transaction.atomic
	def patch(self,request,*args, **kwargs):
		"""Peform the password reset.
		
		Decorators:
			transaction.atomic
		
		Request Data:
			password (string): Email address of registered user

		Status:
			200: Success
			400: No password provided
			401: Link is no longer valid or has been used
			404: Reset code not found   
		"""

		# get key from uri
		key = self.kwargs.get('key')

		# check key for validity
		response = self.validate_key(key)

		# if received Response, invalid key
		if isinstance(response, Response):
			# return the response
			return response
		# otherwise, we have a ResetPasswordRequest
		else:
			instance = response

		# return error if user did not supply a password
		if request.data.get('password', None) == None or request.data.get('password') == "":
			return Response({
					'status': 'fail',
					'message': 'No password set.'
				}, status=status.HTTP_400_BAD_REQUEST);
 

		# set the password
		user = instance.user
		user.set_password(request.data.get('password'))
		user.save()

		instance.used = timezone.now()
		instance.status = ResetPasswordRequest.USED
		instance.save()


		return Response({
				'status': 'success',
				'message': 'Password has been changed.'
			}, status=status.HTTP_200_OK);


	def validate_key(self,key):
		""" Check a key for validity.

		Checks that key exists, has not been disabled, and is not expired.

		Returns:
			Response object for invalid keys
			ResetPasswordRequest for valid keys
		"""

		# get user from password reset items
		try:
			instance = ResetPasswordRequest.objects.get(key=key)
		except:
			# return 404 if no matching reset request
			return Response({
					'status': 'fail',
					'message': 'This password reset key is not valid.'
				}, status=status.HTTP_404_NOT_FOUND);

		# return error if link has been used already or is disabled
		if instance.status == ResetPasswordRequest.USED:
			return Response({
					'status': 'fail',
					'message': 'This password reset key has expired.'
				}, status=status.HTTP_410_GONE);

		elif instance.status == ResetPasswordRequest.DISABLED:
			return Response({
					'status': 'used',
					'message': 'This password reset key has been disabled.'
				}, status=status.HTTP_403_FORBIDDEN);

		# return error if link has expired
		elif instance.created < timezone.now() - timedelta(minutes=AUTH.get('RESET_PASSWORD_EXPIRES')):
			return Response({
					'status': 'fail',
					'message': 'This password reset key has expired.'
				}, status=status.HTTP_410_GONE);

		# check that user is not disabled
		elif instance.user.status == User.DISABLED:
			return Response({
					'status': 'fail',
					'message': 'This account has been disabled.',
				},  status=status.HTTP_403_FORBIDDEN);

		else:
			return instance

	def dispatch_email(self, instance, request):
		""" Send an email to the user containing a link to reset their password.

		Args:
			instance (object): The ResetPasswordRequest instance
			request (object): The request object
		"""

		# create the link
		current_site = get_current_site(request)
		protocol = 'https' if request.is_secure() else 'http' 
		link = '{}://{}/reset-password/{}'.format(protocol, current_site.domain, instance.key)

		# dispatch the email
		send_mail(
			'Password Reset',
			'Reset your password by following this link \n\n{}\n\n-'.format(link),
			'test@example.com',
			[instance.user.email],
			fail_silently=False
			)

class ValidateEmailView(views.APIView):
	""" Determine if an email is already registered with a user account."""


	def get(self,request):

		email = request.GET.get('email')
		userid = request.GET.get('userid')

		# check for existing account with given email
		duplicates = User.objects.filter(email=email)

		# exclude the account already registered with this email
		if userid:
			duplicates = duplicates.exclude(id=userid)

		results = duplicates.all()
		if len(results) > 0:
			return Response({
				'valid': False,
				'message': 'This email is already in use.'
			}, status=status.HTTP_200_OK)
		else:
			return Response({
				'valid': True
			}, status=status.HTTP_200_OK)



from rest_framework import serializers

from django.contrib.auth import get_user_model

from agape.serializers import DynamicSerializer
from .models import UserManager
from .settings import AUTH

User = get_user_model()

from copy import deepcopy


class UserSerializer( DynamicSerializer ):
    password = serializers.CharField(write_only=True, required=False)
    confirm_password = serializers.CharField(write_only=True, required=False)
    email = serializers.CharField(required=False)
    # name  = serializers.CharField(required=False)

    class Meta:
        model = User
        
        fields = [
            'id', 
            'email', 
            # 'name',
            'password', 
            'confirm_password', 
            'status',
            'is_superuser',
            'last_login',
            'display_name'
        ]

        read_only_fields = ['id', 'display_name' ]

        # add the username field if it is not already included as the "email"
        if not AUTH['USERNAME_FIELD'] == 'email':
            fields.append( AUTH['USERNAME_FIELD'] )
            



        



    # def create(self, validated_data):


    #     validated_data = deepcopy(validated_data)
    #     password = validated_data.pop('password', None)

    #     # create the instance
    #     instance = User( **validated_data )

    #     # set the password
    #     instance.set_password( password )

    #     # save
    #     instance.save()            

    #     return instance


    def update(self, instance, validated_data):

        # make a copy of the validated data 
        # validated_data = deepcopy(validated_data)

        # remove the password from the validated database
        password = validated_data.pop('password', None)

        # update the values on the instance
        super().update( instance, validated_data )

        # set the password on the instance if a password was supplied
        if password != None:
            instance.set_password( password )
            instance.save()

        return instance



from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin
from django.contrib.auth import get_user_model
from django.db import models
from .settings import AUTH

from django.contrib.sites.shortcuts import get_current_site


# for random string generation
import string
import random

# for dispatching emails after creating new activation and password reset links
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.core.mail import send_mail

from agape.entity import entity

from agape.models import DynamicModel

class UserManager(BaseUserManager):
    """
    A custom user manager to deal with emails as unique identifiers for
    auth instead of usernames. This replaces the default "UserManager".
    """
    def create_user(self, **args ):
        """
        Creates and saves a User with the given email and password.
        """


        if not args.get( AUTH['USERNAME_FIELD'], None ):
            raise ValueError('The username must be set.')

        if not args.get( 'password', None ):
            raise ValueError('The password must be set.')

        # normalize email field
        if AUTH['USERNAME_FIELD'] == 'email':
            args[ AUTH['USERNAME_FIELD'] ] = self.normalize_email( args[ AUTH['USERNAME_FIELD'] ] )

        # get the password
        password = args.pop('password', None)

        # get the User set in the settings file
        UserModel = get_user_model()

        # set the user
        user = UserModel.objects.create( **args )
        user.set_password(password)
        user.save() 
        return user
    

    def create_superuser(self, **args ):
        args.setdefault('is_superuser',True)
        args.setdefault('status',1)

        return self.create_user(**args)

    def create( self, **fields ):

        set_password = False
        
        if 'password' in fields:
            set_password = True
            password = fields.pop('password', None)

        instance = super().create(**fields)

        if set_password:
            instance.set_password( password )
            instance.save()

        return instance





def email_required():

    return ( AUTH['USERNAME_FIELD'] == 'email'  or AUTH['EMAIL_REQUIRED'] )

def username_required():

    return  AUTH['USERNAME_FIELD'] == 'username' 

@entity
class User(DynamicModel, AbstractBaseUser, PermissionsMixin):
    """
    A custom user class which uses emails as unique identifiers instead
    of usernames. This replaces the default "User" supplied by Django.
    """
    DISABLED = -1
    INACTIVE = 0
    ACTIVE = 1
    STATUS_CHOICES = (
        ("", '-- select --'),
        (DISABLED, 'Disabled'),
        (INACTIVE, 'Inactive'),
        (ACTIVE, 'Active')
    )

    entity = "user"
    
    email        = models.EmailField( unique=True, blank=not email_required(), null=not email_required() )
    # name         = models.CharField( max_length=64, blank=True, null=True )
    username     = models.CharField( max_length=32, unique=username_required(), blank=not username_required(), null=not username_required() )

    is_superuser = models.BooleanField(default=False,help_text="Designates whether user can log into the backend.")
    status       = models.SmallIntegerField(choices=STATUS_CHOICES,blank=False,null=False,default=0 )
    last_login   = models.DateTimeField(blank=True,null=True)
    
    created      = models.DateTimeField(auto_now_add=True,blank=True,null=True)
    created_by   = models.ForeignKey('User', on_delete=models.PROTECT, blank=True, null=True, related_name="created_user")
    modified     = models.DateTimeField(auto_now=True,blank=True,null=True)
    modified_by  = models.ForeignKey('User', on_delete=models.PROTECT, blank=True, null=True, related_name="modified_user")    

    USERNAME_FIELD = AUTH['USERNAME_FIELD']
    REQUIRED_FIELDS = AUTH['REQUIRED_FIELDS']
    objects = UserManager()
    
    @property
    def is_admin(self):
      return self.is_superuser
    
    @property
    def is_staff(self):
      return self.is_superuser

    @property
    def is_active(self):
        return True if self.status == 1 else False;

    @property
    def display_name(self):
        
        if self.username:
            return self.username

        else:
            return self.email
    
    
    def get_short_name(self):
        self.email;
    
    def get_long_name(self):
        self.email;
    
    def __unicode__(self):
        return 'user:{}'.format( self.display_name )

    def __str__(self):
        return 'user:{}'.format( self.display_name )


    def clean_email(self):
        return self.cleaned_data['email'] or None



class UserActivation(models.Model):
    """ Provide activation links for new users.
    
    Generates a random key which is used to activate a new user account. Emails can
    then be sent to users with a link containing the key. Once this link has been
    followed the account will be set to active.
    
    Extends:
        models.Model
    
    Variables:
        user (models.ForeignKey): User account to activate
        key  (models.StringField): Randomly generated string of eight uppercase letters and numbers
        created (models.DateTimeField): Time activation link was created
        activated (models.DateTimeField): Time account was activated
    """

    entity = "user-activation"

    def generate_key():
        """Generate a random string of 8 uppercase letters and numbers

        Returns:
            string: Random string of 8 uppercase letters and numbers

        """
        return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8))

    user      = models.ForeignKey('User', on_delete=models.CASCADE, blank=False, null=False)    
    key       = models.CharField(max_length=8,unique=True,blank=False,null=False,default=generate_key)
    created   = models.DateTimeField(auto_now_add=True,blank=True,null=False)
    activated = models.DateTimeField(blank=True,null=True)



    def send_activation_email( self, request ):
        """ Send the self email to the user """


        current_site = get_current_site(request)
        protocol     = 'https' if request.is_secure() else 'http' 

        # create an self link using information from the 'sites' app
        def create_activation_link():

            link = '{}://{}/activate/{}'
            return link.format( protocol, current_site.domain, self.key )


        # get the from email address from the settings, if there is no domain
        # specified then attach the domain based on information from the 'sites' app
        def get_from_email():

            if not AUTH['FROM_EMAIL_ADDRESS']:
                raise Exception("No from email in settings file. Set the AUTH['FROM_EMAIL_ADDRESS'] value. ")

            elif '@' in AUTH['FROM_EMAIL_ADDRESS']:
                return AUTH['FROM_EMAIL_ADDRESS']

            else:
                return '@'.join( (AUTH['FROM_EMAIL_ADDRESS'], current_site.domain) )


        # create the email
        link = create_activation_link()
        from_email = get_from_email()
        subject = 'Activate your account at {}'
        body    = 'Please activate your account by following this link \n\n{}\n\n'

        try:
            send_mail(
                    subject.format( current_site.name ),
                    body.format( link ),
                    from_email,
                    [ self.user.email ],
                    fail_silently=False
                )
        except Exception as error:

            # raise Exception(
            #    "Error sending email in send_activation_email function: "
            # )
            # TODO: Log this error somewhere! 
            # We need to know if the self email is failing.
            print( "Error sending email in send_activation_email function: ")
            print( error )



class ResetPasswordRequest(models.Model):
    """ Provide password reset links for new users.
    
    Generates a random key which is used to reset user account passwords. Emails can
    then be sent to users with a link containing the key. Once this link has been
    followed the user can set a new password.
    
    Extends:
        models.Model
    
    Variables:
        user (models.ForeignKey): User account to activate
        key  (models.StringField): Randomly generated string of 16 uppercase letters and numbers
        created (models.DateTimeField): Time reset link was created
        activated (models.DateTimeField): Time reset link was followed
    """

    entity = "reset-password-request"

    def generate_key():
        """Generate a random string of 16 uppercase letters and numbers

        Returns:
            string: Random string of 16 uppercase letters and numbers

        """
        return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(16))


    DISABLED = -1
    ENABLED = 1
    USED = 2
    STATUS_CHOICES = (
        ("", '-- select --'),
        (DISABLED, 'Disabled'),
        (ENABLED, 'Active'),
        (USED, 'Used'),
    )

    user      = models.ForeignKey('User', on_delete=models.CASCADE, blank=False, null=False) 
    key       = models.CharField(max_length=16,unique=True,blank=False,null=False,default=generate_key)
    created   = models.DateTimeField(auto_now_add=True,blank=True,null=False)
    status    = models.SmallIntegerField(choices=STATUS_CHOICES,blank=False,null=False,default=1)
    used      = models.DateTimeField(blank=True,null=True)



# enable app initialization events by using the AgapeAppConfig
from agape.django.config import DjangoAppConfig 

class AppConfig( DjangoAppConfig ):
	name = "agape.auth"
	label = "agapeAuth"

# This line prevents this error:  django.core.exceptions.ImproperlyConfigured: Application labels aren't unique, duplicates: auth
default_app_config = 'agape.auth.AppConfig'





from agape.signals import connect
from agape.auth.settings import AUTH
from agape.errors import ErrorResponse

from django.core.mail import send_mail
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site

from django.contrib.auth import authenticate



# Enumerated option values
USER_STATUS_ENABLED = 1;
USER_STATUS_DISABLED = 0;


# get the name of the username field from the configuration
def username_field():
	return  AUTH['USERNAME_FIELD']

# get a dictionary with username key/value
def username_opt(value):
	return { AUTH['USERNAME_FIELD'] : value } 

# get the value of the 
def username_value(instance):
	return getattr(instance, AUTH['USERNAME_FIELD'] )



@connect('user.create:before')
def on_user_create_before_check_unique_username( event,  scope ):
	""" Check that the username field is unique. """
	User = get_user_model()

	username_value = scope.data.get( username_field() )

	existing 	   = User.objects.filter( **username_opt( username_value ) )

	if existing.count() > 0:

		message  = 'This {} is already in use.'

		scope.error = ErrorResponse('409', message.format( username_field() ) )

		raise scope.error



# @connect('user.create(register):before')
@connect('user.register.create:before')
def on_user_create_before_set_status( event,  scope ):
	""" Set the status of the user based on user configuration. Status will
	be enabled if accounts do not require activation."""

	#TODO: if scope.premise == 'register':

	scope.data['status'] = USER_STATUS_DISABLED \
								if AUTH['REQUIRE_ACTIVATION'] \
								else USER_STATUS_ENABLED




@connect('user.register.create:success')
def on_user_create_success( event, scope ):

	from agape.auth.models import UserActivation

	# if account requires activation
	if AUTH['REQUIRE_ACTIVATION']:

		activation = UserActivation.objects.create( user=scope.instance )

		send_activation_email( activation, scope.request )




def send_activation_email( activation, request ):
	""" Send the activation email to the user """


	current_site = get_current_site(request)
	protocol     = 'https' if request.is_secure() else 'http' 

	# create an activation link using information from the 'sites' app
	def create_activation_link():

		link = '{}://{}/activate/{}'
		return link.format( protocol, current_site.domain, activation.key )


	# get the from email address from the settings, if there is no domain
	# specified then attach the domain based on information from the 'sites' app
	def get_from_email():

		if not AUTH['FROM_EMAIL_ADDRESS']:
			raise Exception("No from email in settings file. Set the AUTH['FROM_EMAIL_ADDRESS'] value. ")

		elif '@' in AUTH['FROM_EMAIL_ADDRESS']:
			return AUTH['FROM_EMAIL_ADDRESS']

		else:
			return '@'.join( (AUTH['FROM_EMAIL_ADDRESS'], current_site.domain) )


	# create the email
	link = create_activation_link()
	from_email = get_from_email()
	subject = 'Activate your account at {}'
	body 	= 'Please activate your account by following this link \n\n{}\n\n'

	try:
		send_mail(
				subject.format( current_site.name ),
				body.format( link ),
				from_email,
				[ activation.user.email ],
				fail_silently=False
			)
	except Exception as error:

		# raise Exception(
		# 	 "Error sending email in send_activation_email function: "
		# )
		# TODO: Log this error somewhere! 
		# We need to know if the activation email is failing.
		print( "Error sending email in send_activation_email function: ")
		print( error )



@connect('user.update:before')
def on_user_update_authenticate_on_modify_credentials( event, scope ):

	scope.modifying_current_user = str(scope.pk) == str( scope.request.user.id )

	if AUTH['REQUIRE_PASSWORD_ON_ACCOUNT_MODIFICATIONS'] and scope.modifying_current_user:

		# error no password supplied
		if not 'confirm_password' in scope.data:

			raise ErrorResponse('401', 'Confirm your password to make account changes.')

		# authenticate the current user with the confirm_password option
		user = authenticate( 
			# this line retrieves the "username" field and value using the app configuration 
			**username_opt( username_value(scope.request.user) ), 
			password=scope.data.get('confirm_password')
		) 

		if user is None:
			raise ErrorResponse('401', 'Password incorrect.')

		elif user.is_active == False:
			raise ErrorResponse('401', 'Account disabled.')

@connect('user.update:before')
def on_user_update_before_get_new_password( event, scope ):

	if 'password' in scope.data:
		scope.new_password = scope.data.get('password', None)
		scope.data.pop('password')
	else:
		scope.new_password = None



@connect('user.update:success')
def on_user_update_success_set_password( event, scope ):

	if scope.new_password:
	    scope.instance.set_password( scope.new_password )
	    scope.instance.save()
	    

@connect('user.update:success')
def on_user_update_success( event, scope ):
	if scope.new_password and hasattr( scope, 'modifiying_current_user' ):
		update_session_auth_hash( scope.request, scope.instance )




from django.test import TestCase
from rest_framework.test import APIRequestFactory
from rest_framework.test import APIClient
import json
from django.utils import timezone
from datetime import timedelta

# Create your tests here.
from .models import User, UserManager, UserActivation, ResetPasswordRequest
from .serializers import UserSerializer

from .settings import AUTH, defaults as AUTH_DEFAULTS


def username_opt(value):
    return { AUTH['USERNAME_FIELD'] : value } 

def username_attr(instance):
    return getattr(instance, AUTH['USERNAME_FIELD'] )


import copy
from agape.test import AgapeTests, ApiMixin, ModelTests,  SerializerTests, TestCaseMixin


USER_MODEL = User
USER_SERIALIZER = UserSerializer
USER_ENTITY_NAME = 'user'
USER_API_END_POINT = '/api/users/'
AUTH_API_END_POINT = '/api/auth/'
REGISTER_API_END_POINT = "/api/register"
RECOVERY_API_END_POINT = "/api/auth/recovery/"


USER_CREATE_DATA = [

    {
        AUTH['USERNAME_FIELD']: 'test@example.com',
        'password': 'testpass',
        # 'name': 'Test User'
    }

]

USER_UPDATE_DATA = [

    {
        AUTH['USERNAME_FIELD']: 'info@codewise.io',
    }

]


class UserTestMixin( TestCaseMixin ):

    def sharedSetUp(self):
        self.resetSettings()

        self.model              = USER_MODEL
        self.serializer_class   = USER_SERIALIZER
        self.entity             = USER_ENTITY_NAME
        self.api_end_point      = USER_API_END_POINT
        self.auth_end_point     = AUTH_API_END_POINT
        self.register_end_point = REGISTER_API_END_POINT
        self.recovery_end_point = RECOVERY_API_END_POINT

        self.create_data        = copy.deepcopy(USER_CREATE_DATA)
        self.expect_data        = copy.deepcopy(USER_CREATE_DATA)
        for record in self.expect_data:
            record.pop('password')
            
        self.update_data        = copy.deepcopy(USER_UPDATE_DATA)
        self.update_expect_data = copy.deepcopy(USER_UPDATE_DATA)

    def resetSettings(self):
        # reset configuration to default settings
        for key, value in AUTH_DEFAULTS.items():
            AUTH[key] = AUTH.get(key, AUTH_DEFAULTS[key] )

    def register_user( self, data ):
        response = self.client.post( self.register_end_point, data )
        self.assertEqual(response.status_code, 201, "Registered user")
        return response

class ModelTestCase( UserTestMixin, ModelTests, TestCase ):
    pass








# TODO, DELETE THESE TESTS, REPLACE WITH TESTS SPECIFICALLY FOR ACCOUNT ACTIVATION/RESET
class UserTestCase(UserTestMixin, TestCase):

    def setUp(self):
        self.resetSettings()
        user = User( **username_opt("test@example.com") )
        user.set_password("password")
        user.save()

    def test_user_account(self):
        """User account is created and retrieved."""
        user = User.objects.get( **username_opt("test@example.com") )
        self.assertEqual( username_attr(user), 'test@example.com')

    def test_account_activation(self):
        user = User.objects.get( **username_opt("test@example.com") )
        instance = UserActivation.objects.create(user=user)
        self.assertTrue(instance.key, "Created activation key")

    def test_password_reset(self):
        user = User.objects.get( **username_opt("test@example.com") )
        instance = ResetPasswordRequest.objects.create(user=user)
        self.assertTrue(instance.key, "Created activation key")   


class UserManagerTestCase(TestCase):
    
    def test_create_user(self):
        user_manager = UserManager()
        user = user_manager.create_user(email='test@example.com',password='testingpassword')

        user = User.objects.get( **username_opt("test@example.com") )
        self.assertEqual( username_attr(user),  'test@example.com')

    def test_create_superuser(self):
        user_manager = UserManager()
        user = user_manager.create_superuser(email='admin@example.com',password='testingpassword')  

        user = User.objects.get( **username_opt("admin@example.com") )
        self.assertEqual(username_attr(user),  'admin@example.com')     






class UserSerializerTestCase( UserTestMixin, SerializerTests, TestCase):


    def serializedSetUp(self):

        super().serializedSetUp()

        # remove the password from the expected serialized data
        for expect_data in ( self.serialized_expect_data, self.serialized_update_expect_data ):
            for record in expect_data:
                record.pop('password', None)



    def test_change_password_via_serializer( self ):

        instance = self.model.objects.create( **self.create_data[0] )
        instance_id = instance.id
        original_password = instance.password

        serializer = self.serializer_class( instance, data={'password':'newpass'}, partial=True )

        if serializer.is_valid( raise_exception=True) :
            serializer.update( instance, serializer.validated_data )

        # just test that the password has changed, leave it to django to test
        # what the password actually changed to
        instance = None
        instance = self.model.objects.get( pk=instance_id )
        self.assertTrue( instance.password != original_password )



class AccountRecoveryTestCase( UserTestMixin, ApiMixin, AgapeTests, TestCase ):

    def setUp(self):
        self.client =  APIClient()
        self.sharedSetUp()
        self.resetSettings()

        self.api_end_point = self.recovery_end_point

    def test_password_reset(self):
        self.register_user( self.create_data[0] )

        # activate user
        user = User.objects.get( email=self.create_data[0]['email'] )
        user.status = 1;
        user.save()

        # create password reset request
        response = self.post( {'email': 'test@example.com', 'reset': True} )
        self.assertEqual(response.status_code, 200, "Reset password request created")

        # check for existence of database entry
        instance = ResetPasswordRequest.objects.filter(user=user).order_by('-id')[0]
        self.assertTrue(instance, "Reset password request object exists")

        # check password reset link is valid
        response = self.client.get('/api/auth/recovery/{}/'.format(instance.key))
        self.assertEqual(response.status_code, 200, "Validate password reset link ok")   

        # try creating another password reset and verify that they have the same key
        response = self.client.post('/api/auth/recovery/', {'email': 'test@example.com', 'reset': True})
        duplicate = ResetPasswordRequest.objects.filter(user=user).order_by('-id')[0]
        self.assertEqual(instance.key, duplicate.key, "Used the same password reset request")   

        # update the original reset request to be over an hour old
        duplicate = None
        time_threshold = timezone.now() - timedelta(minutes=AUTH.get('RESET_PASSWORD_EXPIRES')+1)
        instance.created = time_threshold
        instance.save()

        # check password reset link is now expired
        response = self.client.get('/api/auth/recovery/{}/'.format(instance.key))
        self.assertEqual(response.status_code, 410, "Password link has expired")

        # reset the password using the expired link (should fail)
        response = self.client.patch('/api/auth/recovery/{}/'.format(instance.key), json.dumps({'password': 'newpass'}), content_type='application/json')
        self.assertEqual(response.status_code, 410, "Do not reset password on expired link")


        # create another password reset and verify that they have different keys
        response = self.client.post('/api/auth/recovery/', {'email': 'test@example.com', 'reset': True})
        duplicate = ResetPasswordRequest.objects.filter(user=user).order_by('-id')[0]
        self.assertTrue(not instance.key == duplicate.key, "Created new password reset request") 

        # reset the password using the new link (should pass)
        response = self.client.patch('/api/auth/recovery/{}/'.format(duplicate.key), json.dumps({'password': 'newpass'}), content_type='application/json')
        self.assertEqual(response.status_code, 200, "Reset password on valid link")

        # check password reset link is now expired (has been used)
        response = self.client.get('/api/auth/recovery/{}/'.format(instance.key))
        self.assertEqual(response.status_code, 410, "Password link has been used")

        # reset the password using the new link again (should fail)  
        response = self.client.patch('/api/auth/recovery/{}/'.format(duplicate.key), json.dumps({'password': 'newpass'}), content_type='application/json')
        self.assertEqual(response.status_code, 410, "Do not reset on used link")

        # create a new password reset request, validate that it is a unique key
        used = duplicate

        # create another password reset and verify that they have different keys
        response = self.client.post('/api/auth/recovery/', {'email': 'test@example.com', 'reset': True})
        another = ResetPasswordRequest.objects.filter(user=user).order_by('-id')[0]
        self.assertTrue(not used.key == another.key, "Created new password reset request") 

        # submit the change request with no password (should fail)
        response = self.client.patch('/api/auth/recovery/{}/'.format(another.key), json.dumps({'password': ''}), content_type='application/json')
        self.assertEqual(response.status_code, 400, "No password supplied error")

        # set the new link to disabled
        another.status = ResetPasswordRequest.DISABLED
        another.save()

        # check password reset link is now disabled
        response = self.client.get('/api/auth/recovery/{}/'.format(another.key))
        self.assertEqual(response.status_code, 403, "Link has been disabled")

        # try resetting password on disbaled link
        response = self.client.patch('/api/auth/recovery/{}/'.format(another.key), json.dumps({'password': 'newpass'}), content_type='application/json')
        self.assertEqual(response.status_code, 403, "Link has been disabled")

        # set the new link to enabled
        another.status = ResetPasswordRequest.ENABLED
        another.save()

        # set the user to disabled
        another.user.status = User.DISABLED
        another.user.save()

        # check password reset link is now disabled
        response = self.client.get('/api/auth/recovery/{}/'.format(another.key))
        self.assertEqual(response.status_code, 403, "User has been disabled")

        # try resetting password on disbaled link
        response = self.client.patch('/api/auth/recovery/{}/'.format(another.key), json.dumps({'password': 'newpass'}), content_type='application/json')
        self.assertEqual(response.status_code, 403, "User has been disabled")  



class APITestCase(UserTestMixin, ApiMixin, AgapeTests, TestCase):

    def setUp(self):
        self.client =  APIClient()
        self.sharedSetUp()
        self.resetSettings()

    def test_login_using_username_field(self):

        # register user
        instance = User.objects.create( **{'password':'testing', 'email':'test@email.com', 'status': 1})

        # attempt to login with credentials (should fail because user is inactive)
        response = self.client.post(self.auth_end_point, {'username': 'test@email.com', 'password':'testing'})
        self.assertEqual(response.status_code, 200, "Logged in")


    def test_activate_disabled_account_error(self):

        self.register_user( self.create_data[0] )

        # disable user
        user = User.objects.get( email=self.create_data[0]['email'] )
        user.status = -1;
        user.save();

        user = User.objects.get( email=self.create_data[0]['email'] )
        self.assertEqual(user.status, -1, "New user is disabled")

        # check if activation link created
        activation =  UserActivation.objects.get(user_id=user.id)
        self.assertTrue( activation, "User activation key generated")
        self.assertTrue( not activation.activated, "Activation has not been used");

        # follow activation link (expect an error)
        response = self.client.patch(self.auth_end_point, json.dumps({'key':activation.key}), content_type='application/json')
        self.assertEqual(response.status_code, 400, "Can not activate disabled user")


  


    def test_validate_email(self):

        # check if email is available ( yes available )
        response = self.client.get('/api/auth/email/?email=test@example.com')
        self.assertEqual(response.data.get('valid'), True, "No conflict")

        # register user
        self.register_user( self.create_data[0] )   

        user = User.objects.get( email=self.create_data[0]['email'] )

        # check if email is available ( no, in use )
        response = self.client.get('/api/auth/email/?email=test@example.com')
        self.assertEqual( response.data.get('valid'), False, "Conflict" )

        # check if email is available ( yes, do not flag users email as invalid )
        response = self.client.get('/api/auth/email/?email=test@example.com&userid={}'.format(user.id))
        self.assertEqual(response.data.get('valid'), True, "No conflict")




    def test_change_credentials_password_required_and_correct(self):

        cdata = self.create_data[0]

        self.register_user( cdata )

        # create a user
        email=cdata['email']

        user = User.objects.get( email=cdata['email'] )
        user.status = 1;
        user.save()

        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Logged in")

        userid = response.data['user'].get('id')
        response = self.patch(userid, {'password':'newpass','confirm_password':cdata['password']} )
        self.assertEqual(response.status_code, 200, "Updated password")

        response = self.authenticate( email=cdata['email'], password="newpass" )
        self.assertEqual(response.status_code, 200, "Logged in")

        # update email 
        response = self.patch(userid, {'email':'updated@example.com','confirm_password':'newpass'})
        self.assertEqual(response.status_code, 200, "Updated email")   
        
        # login with new credentials
        response = self.authenticate( email='updated@example.com', password="newpass" )
        self.assertEqual(response.status_code, 200, "Signed in with new credentials")



    def test_change_credentials_password_required_not_supplied_throws_exception(self):

        # register a user
        cdata    = self.create_data[0]
        response = self.register_user( cdata )
        userid   = response.data['id']

        # set user as active
        user = User.objects.get(pk=userid)
        user.status = 1;
        user.save()

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")

        # update password - this needs to fail, confirm_password not supplied
        userid = response.data['user'].get('id')    
        response = self.patch( userid, {'password':'newpass'} )
        self.assertEqual(response.status_code, 401, "Not authorized")    



    def test_change_credentials_reauthentication_not_required(self):

        AUTH['REQUIRE_PASSWORD_ON_ACCOUNT_MODIFICATIONS'] = False

        # register a user
        cdata    = self.create_data[0]
        response = self.register_user( cdata )
        userid   = response.data['id']

        # set user as active
        user = User.objects.get(pk=userid)
        user.status = 1;
        user.save()

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")

        # change password without reauthentication
        userid = response.data['user'].get('id')    
        response = self.patch( userid, {'password':'newpass'} )  

        AUTH['REQUIRE_PASSWORD_ON_ACCOUNT_MODIFICATIONS'] = True




    def test_last_login(self):

        AUTH['REQUIRE_ACTIVATION'] = False

        # register a user
        cdata    = self.create_data[0]
        response = self.register_user( cdata )
        userid   = response.data['id']

        # verify last login is blank
        user = User.objects.get( pk=userid )
        self.assertEqual(user.last_login, None, "User has never logged in")

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")

        # verify last login is not blank
        user = User.objects.get( pk=userid )
        self.assertTrue(user.last_login, "User has logged in")

        AUTH['REQUIRE_ACTIVATION'] = True

  
    def test_list(self):

        # register a user
        cdata    = self.create_data[0]
        response = self.register_user( cdata )
        userid   = response.data['id']

        # activate user
        user = User.objects.get(email="test@example.com")
        user.status = 1;
        user.save()

        # create users
        for n in range(1,6):
            User.objects.create( email='user{}@example.com'.format(n), password="testing")
        
        # list user (without permissions)
        response = self.client.get(self.api_end_point)
        self.assertEqual(response.status_code, 403, "List users unauthorized")  

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")

        # list user (without permissions)
        response = self.client.get(self.api_end_point)
        self.assertEqual(response.status_code, 403, "List users unauthorized")   

        # add permissions to list users
        user.is_superuser = True
        user.save()   

        # list user (with super admin permissions)
        response = self.client.get(self.api_end_point)
        self.assertEqual(response.status_code, 200, "List users as super admin")  




    def test_permissions(self):

        # register a user
        cdata    = self.create_data[0]
        response = self.register_user( cdata )
        userid   = response.data['id']

        # activate user
        user = User.objects.get(email="test@example.com")
        user.status = 1;
        user.save()


        # retrieve user (unauthorized)
        response = self.get( userid )
        self.assertEqual(response.status_code, 403, "Retrieve user unauthorized")     

        # login
        response = self.authenticate( email=cdata['email'], password=cdata['password'] )
        self.assertEqual(response.status_code, 200, "Signed in")
        
        # retrieve user (authorized)
        response = self.get( userid )
        self.assertEqual(response.status_code, 200, "Retrieved user data")    



                
class RegisterAPITestCase ( TestCaseMixin, ApiMixin, AgapeTests, TestCase ):

    def setUp(self):
        
        self.sharedSetUp()
        self.createRequisiteObjects()
        self.resetSettings()

    def sharedSetUp(self):
        self.model              = USER_MODEL
        self.serializer_class   = USER_SERIALIZER
        self.entity             = USER_ENTITY_NAME
        self.api_end_point      = '/api/register/'
        self.auth_end_point     = AUTH_API_END_POINT

        self.create_data        = copy.deepcopy(USER_CREATE_DATA)
        self.expect_data        = copy.deepcopy(USER_CREATE_DATA)
        self.update_data        = copy.deepcopy(USER_UPDATE_DATA)
        self.update_expect_data = copy.deepcopy(USER_UPDATE_DATA)

    def resetSettings(self):
        # reset configuration to default settings
        for key, value in AUTH_DEFAULTS.items():
            AUTH[key] = AUTH.get(key, AUTH_DEFAULTS[key] )


    def test_user_registration(self):

        self.resetSettings()

        # register user
        response = self.post( 
            {'email': 'test@example.com', 'password':'testing'}
        )
        self.assertEqual(response.status_code, 201, "Created new user")
        
        # get user, verify that user is not active
        user = User.objects.get(email="test@example.com")
        self.assertEqual(user.status, 0, "New user is inactive")

        # attempt to login with credentials (should fail because user is inactive)
        response = self.client.post(self.auth_end_point, {'email': 'test@example.com', 'password':'testing'})
        self.assertEqual(response.status_code, 401, "Could not login to inactive account")

        # check if activation link created
        activation =  UserActivation.objects.get(user_id=user.id)
        self.assertTrue(activation, "User activation key generated")
        self.assertTrue(not activation.activated, "Activation has not been used");

        # follow activation link
        response = self.client.patch(self.auth_end_point, json.dumps({'key':activation.key}), content_type='application/json')
        self.assertEqual(response.status_code, 200, "User has been activated")

        # check if activation link is used
        activation =  UserActivation.objects.get(key=activation.key)
        self.assertTrue(activation.activated, "Activation key has been used");

        # check if user is now active
        user = User.objects.get(email="test@example.com")
        self.assertEqual(user.status, 1, "User is now active")


        # attempt to login with credentials (should pass because account has been activated)
        response = self.client.post(self.auth_end_point, {'email': 'test@example.com', 'password':'testing'})
        self.assertEqual(response.status_code, 200, "Sign in successful")
        self.assertEqual(response.data['user'].get('email'), "test@example.com")
        self.assertTrue(response.data.get('token'), "Received auth token.")

        # logout
        response = self.client.delete(self.auth_end_point)
        self.assertEqual(response.status_code, 204, "Sign out successful")


    def test_register_duplicate_email_error(self):
        # register user
        response = self.client.post(self.api_end_point, {'email': 'test@example.com', 'password':'testing'})
        self.assertEqual(response.status_code, 201, "Created new user")

        response = self.client.post(self.api_end_point, {'email': 'test@example.com', 'password':'testing'})
        self.assertEqual(response.status_code, 409, "Could not register duplicate account")

    def test_register_without_user_activation(self):
        AUTH['REQUIRE_ACTIVATION'] = False

        # register user
        response = self.client.post(
            self.api_end_point, {'email': 'test@example.com', 'password':'testing'}
        )
        self.assertEqual(response.status_code, 201, "Created new user")


        # get user, verify that user is not active
        user = User.objects.get(email="test@example.com")
        self.assertEqual(user.status, 1, "New user active")

        AUTH['REQUIRE_ACTIVATION'] = True




# test that event is triggered when the agape.auth app 
# is ready
from agape.signals import connect
appready = False

@connect('django.app.ready')
def notice( event, scope ):
    print ( event )
    print('HEEEEEEEEEEEEEEEEEEEEEEEEEYYYYYYYYYYYYYY')
    appready = True


class AuthEventsTestCase( TestCase ):

    def setUp(self):
        pass

    def test_app_ready_notice_called( self ):

        self.assertTrue( True )


from django.contrib import admin


from .models import User,UserActivation,ResetPasswordRequest

admin.site.register(User)
admin.site.register(UserActivation)
admin.site.register(ResetPasswordRequest)